"""Basic Spectrum Analyzer classes intended for expansion with hardware aware classes."""

from qolab.hardware.scpi import SCPIinstr
from qolab.hardware.basic import BasicInstrument
from qolab.data.trace import TraceSetSameX


class SpectrumAnalyzer(BasicInstrument):
    """Basic Spectrum Analyzer class.
    Intended to be expanded in hardware specific classes.
    """

    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config["Device type"] = "Spectrum Analyzer"
        self.config["Device model"] = (
            "Generic Spectrum Analyzer Without Hardware interface"
        )
        self.config["FnamePrefix"] = "spectrum_analyzer"
        self.numberOfChannels = 0
        self.deviceProperties.update(
            {
                "FreqCenter",
                "FreqSpan",
                "FreqStart",
                "FreqStop",
                "RBW",
                "VBW",
                "SweepTime",
            }
        )
        self.channelProperties = {
        }

    # Minimal set of methods to be implemented.
    def getTrace(
        self,
        chNum,
        **kwargs,
    ):
        """Get scope trace with time axis set."""
        raise NotImplementedError("getTrace function is not implemented")

    def getAllTraces(self, channelsList=None, **kwargs):
        """
        Get all traces channels in the channelsList.

        If channelsList is None, go over all physical channels: 1, 2, ...
        channelsList can contain non hardware channels, i.e. functional one
        then they should be called F1, F2, ... etc
        or have hardware scope dependent names to be deciphered in a particular scope class.
        """
        if channelsList is None:
            channelsList = range(1, self.numberOfChannels + 1)
        allTraces = TraceSetSameX("spectrum analyzer traces")
        allTraces.config["tags"]["DAQ"] = self.getConfig()
        for chNum in channelsList:
            allTraces.addTrace(
                self.getTrace(
                    chNum,
                    **kwargs
                )
            )
        return allTraces

    def save(
        self,
        fname=None,
        item_format="e",
        channelsList=None,
        extension="dat.gz",
    ):
        allTraces = self.getAllTraces(
            channelsList=channelsList,
        )
        allTraces.config["item_format"] = item_format
        if fname is None:
            fname = self.getNextDataFile(extension=extension)
        allTraces.save(fname)
        print(f"Data saved to: {fname}")
        return fname

class SpectrumAnalyzerSCPI(SCPIinstr, SpectrumAnalyzer):
    """Basic Spectrum Analyzer class with SCPI interface.
    Intended to be expanded in hardware specific classes.

    Example
    -------
    >>> rm = pyvisa.ResourceManager()
    >>> SpectrumAnalyzerSCPI(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    """

    pass

    def __init__(self, resource, *args, **kwds):
        SCPIinstr.__init__(self, resource)
        SpectrumAnalyzer.__init__(self, *args, **kwds)
        self.config["DeviceId"] = str.strip(self.idn)



from qolab.hardware.basic import BasicInstrument
from ._basic import PowerSupplySCPI


class KeysightE36231A(PowerSupplySCPI):
    """Keysight E36231A power supply"""

    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.resource.read_termination = "\n"
        self.config["Device model"] = "Keysight E36231A"
        self.numberOfChannels = 1
        self.deviceProperties.update({"OpMode"})
        self.channelProperties = {
            "IsOn",
            "Regulation",
            "Vout",
            "Vlimit",
            "Iout",
            "Ilimit",
            "dV",
            "dI",
        }

    def setOpMode(self, val):
        """Sets power supply operation mode, returns OFF|PAR|SER.

        OFF stands for independent channels.
        """
        ALLOWED_CMD = ['OFF', 'PAR', 'SER']
        try:
            if val in ALLOWED_CMD:
                cmnd = f"OUTP:PAIR {val}"
                self.write(cmnd)
            else:
                print(f"[WARNING] Command PSU operation mode {val} is not allowed")
        except pyvisa.errors.VisaIOError as e:
                print(f"[ERROR] GPIB communication error: {e}")

    def setChanOn(self, chNum=1):
        """Power up channel output"""
        self.write(f"OUTP ON,(@{chNum})")

    def setChanOff(self, chNum=1):
        """Power down channel output"""
        self.write(f"OUTP OFF,(@{chNum})")

    @BasicInstrument.tsdb_append
    def getChanIsOn(self, chNum=1):
        """Queries channel output state"""
        qstr = f"OUTP? (@{chNum})"
        rstr = self.query(qstr)
        return bool(float(rstr))

    @BasicInstrument.tsdb_append
    def getChanRegulation(self, chNum=1):
        """Queries channel output regulation

        0 - The output is off and unregulated
        1 - The output is CC (constant current) operating mode
        2 - The output is CV (constant voltage) operating mode
        3 - The output has hardware failure
        """
        qstr = f"STAT:QUES:INST:ISUM{chNum}:COND?"
        rstr = self.query(qstr)
        return int(rstr)

    @BasicInstrument.tsdb_append
    def getChanVout(self, chNum=1):
        qstr = f"MEAS:VOLT? (@{chNum})"
        rstr = self.query(qstr)
        return float(rstr)

    @BasicInstrument.tsdb_append
    def getChanVlimit(self, chNum=1):
        qstr = f"SOUR:VOLT? (@{chNum})"
        rstr = self.query(qstr)
        return float(rstr)

    @BasicInstrument.tsdb_append
    def getChanIout(self, chNum=1):
        qstr = f"MEAS:CURR? (@{chNum})"
        rstr = self.query(qstr)
        return float(rstr)

    def get_out_current(self):
        '''
        Wrapper for `getChanIout()` to comply with PSW25045 
        source's method `get_out_current`.
        '''
        return self.getChanIout()
    
    def get_out_voltage(self):
        '''
        Wrapper for `getChanVout()` to comply with PSW25045 
        source's method `get_out_voltage`.
        '''
        return self.getChanVout()

if __name__ == "__main__":
    import pyvisa

    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    instr = rm.open_resource("USB0::0x2A8D::0x2F02::MY61003701::INSTR")
    ps = KeysightE36231A(instr)
    print("------ Header start -------------")
    print(str.join("\n", ps.getHeader()))
    print("------ Header ends  -------------")


from qolab.hardware.spectrum_analyzer._basic import SpectrumAnalyzerSCPI
from qolab.hardware.scpi import SCPI_PROPERTY
from qolab.data.trace import Trace, TraceXY
import numpy as np


class Agilent_E4405B(SpectrumAnalyzerSCPI):
    """Agilent E4405B Spectrum Analyzer"""

    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.config["Device model"] = "Agilent E4405B"
        self.resource.read_termination = "\n"
        self.numberOfChannels = 3
        self.deviceProperties.update(
            {
                "YAxisScale",
                "YAxisUnit",
                "ReferenceLevel",
                "Attenuation",
                "TraceDataExchangeFromat",
                "SweepNpoints",
            }
        )

    @property
    def idn(self):
        """Agilent E4405B  has non standard call of idn"""
        return self.query("ID?")

    FreqCenter = SCPI_PROPERTY(
        scpi_prfx="SENSE:FREQUENCY:CENTER", ptype=float, doc="Central frequency"
    )
    FreqSpan = SCPI_PROPERTY(
        scpi_prfx="SENSE:FREQUENCY:SPAN", ptype=float, doc="Frequency span"
    )
    FreqStart = SCPI_PROPERTY(
        scpi_prfx="SENSE:FREQUENCY:START",
        ptype=float,
        doc="Start frequency of the span",
    )
    FreqStop = SCPI_PROPERTY(
        scpi_prfx="SENSE:FREQUENCY:STOP", ptype=float, doc="Stop frequency of the span"
    )
    RBW = SCPI_PROPERTY(
        scpi_prfx="SENSE:BANDWIDTH:RESOLUTION", ptype=float, doc="Resolution bandwidth"
    )
    VBW = SCPI_PROPERTY(
        scpi_prfx="SENSE:BANDWIDTH:VIDEO", ptype=float, doc="Video bandwidth"
    )
    SweepTime = SCPI_PROPERTY(
        scpi_prfx="SENSE:SWEEP:TIME", ptype=float, doc="Video bandwidth"
    )
    YAxisScale = SCPI_PROPERTY(
        scpi_prfx=":DISPlAY:WINDOW:TRACE:Y:SCALE:PDIVISION",
        ptype=float,
        doc="Vertical axis scale per division",
    )
    YAxisUnit = SCPI_PROPERTY(
        scpi_prfx=":UNIT:POWER",
        ptype=str,
        doc="Vertical axis unit (DBM|DBMV|DBUV|DBUA|V|W|A)",
    )
    ReferenceLevel = SCPI_PROPERTY(
        scpi_prfx="DISPLAY:WINDOW:TRACE:Y:SCALE:RLEVEL",
        ptype=float,
        doc="Reference level",
    )
    Attenuation = SCPI_PROPERTY(
        scpi_prfx=":SENSE:POWER:RF:ATTenuation",
        ptype=float,
        doc="Attenuation level, default setting is 10 and unit dB",
    )
    TraceDataExchangeFromat = SCPI_PROPERTY(
        scpi_prfx=":FORMAT:TRACE:DATA",
        ptype=str,
        doc="""
        Trace data exchange format (ASCii|INTeger,32|REAL,32|REAL,64|UINTeger,16),
        defalult is ASCII - human readable in the units of the Yscale
        fastest INTEGER,32 which report mdBm units (internal to the spectrum analyzer),
        """,
    )
    SweepNpoints = SCPI_PROPERTY(
        scpi_prfx=":SENSe:SWEep:POINts",
        ptype=int,
        doc="Number of points per sweep trace. Default 401, can be as large as 8192",
    )

    def getXTrace(self):
        # return X trace, default is frequency space
        # but for zero span it is better to use time space
        Npnts = self.SweepNpoints
        # de
        x = Trace("ToBeRedefined")
        x.config["tags"]["Npnts"] = Npnts
        x.config["tags"]["FreqCenter"] = self.FreqCenter
        x.config["tags"]["FreqSpan"] = self.FreqSpan
        x.config["tags"]["SweepTime"] = self.SweepTime
        if 0 != x.config["tags"]["FreqSpan"]:
            x.config["label"] = "Frequency"
            x.config["unit"] = "Hz"
            xval = np.linspace(self.FreqStart, self.FreqStop, Npnts)
        else:
            x.config["label"] = "Time"
            x.config["unit"] = "s"
            xval = np.linspace(0, x.config["tags"]["SweepTime"], Npnts)
        x.values = xval.reshape(xval.size, 1)
        return x

    def getTrace(
        self,
        chNum,
        **kwargs,
    ):
        chName = "Ch"+str(chNum)
        qstr = f':TRACE:DATA? TRACE{chNum};'
        fmt = self.TraceDataExchangeFromat
        if 'ASC,+8' == fmt:
            rstr = self.query(qstr)
            lstr = rstr.split(',')
            n=np.array([float(i) for i in lstr])
        else:
            raise(RuntimeError(f"Do not know how to exchange data formatted as '{fmt}'"))
        tr = TraceXY(chName)
        x = self.getXTrace()
        tr.x = x
        y = Trace("Amplitude")
        y.config["unit"] = self.YAxisUnit
        y.config["tags"]["ReferenceLevel"] = self.ReferenceLevel
        y.config["tags"]["Attenuation"] = self.Attenuation
        y.values = n.reshape(n.size,1)
        tr.y = y
            
        return tr

if __name__ == "__main__":
    import pyvisa

    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    instr = rm.open_resource("visa://10.160.137.40/GPIB0::21::INSTR")

    sa = Agilent_E4405B(instr)
    print("------ Header start -------------")
    print(str.join("\n", sa.getHeader()))
    print("------ Header ends  -------------")
